/****************************************************************************
Gui interface for OLSRD
Use the plugin txtinfo with patch for XML formatting

Patrick Gerber <patrick.gerber@heig-vd.ch>
Copyright (c) 2008 ReDS Institute <reds.heig-vd.ch>
june 2008 - v0.1

Developped for ubidule under the Perplexus project
See www.perplexus.org for more informations

Released under the same licence as OLSR v0.5.5
See www.olsr.org for more informations
****************************************************************************/

#include <QtGui>
#include <QtNetwork>

#include "olsrdInfo.h"
#include <math.h>
#include "arrow.h"
#include <iostream>

#define OLSRD_PORT 2006

OlsrdInfo::OlsrdInfo(QString a_addr, QWidget *parent) : QWidget(parent)
{
	timer = new QTimer();
	buf = new QString();
	topologies = new QMap<QString, QString>;
	posNodes = new QMap<QString,QPointF>;
	nodesInfos = new QMap<QString, QString>;
	setMouseTracking(true);
	this->addr = QString(a_addr);
	anim = 0.0;

    connect(timer, SIGNAL(timeout()), this, SLOT(socket_start()));
    connect(&tcpClient, SIGNAL(connected()), this, SLOT(socket_connected()));
    connect(&tcpClient, SIGNAL(readyRead()), this, SLOT(socket_readyRead()));
    connect(&tcpClient, SIGNAL(disconnected()), this, SLOT(socket_disconnected()));
    connect(&tcpClient, SIGNAL(error(QAbstractSocket::SocketError)),
            this, SLOT(socket_displayError(QAbstractSocket::SocketError)));
    
    setWindowTitle(tr("QOlsrd - Qt Gui for OLSR - ReDS"));
    timer->start(1000);
    socket_start();
}

/*
 Connect the socket to the server (OLSRD)
*/
void OlsrdInfo::socket_start()
{
	//anim += 0.01;
   	buf->clear();	
   	if(addr.isEmpty())
   		tcpClient.connectToHost("127.0.0.1", OLSRD_PORT);
   	else
   		tcpClient.connectToHost(addr, OLSRD_PORT);
}

/*
  The socket is now connected to the OLSR server
  we start to download information 
*/
void OlsrdInfo::socket_connected()
{
	tcpClient.write("/xml\n");
}

/*
  New data is ready to read
  We append new data in the buffer for XMl parser 
*/
void OlsrdInfo::socket_readyRead()
{  
	QString tmp = tcpClient.readAll();
	buf->append(tmp);
}

/*
  Connection with the server as been closed
  We can now parse and display all information
*/
void OlsrdInfo::socket_disconnected()
{
	 QString errorMsg;
	 QDomDocument *xmlParser = new QDomDocument();
	 int errorLine, errorColumn;
	 topologies->clear();
	 
	 //Load received text in the XML parser
	 xmlParser->clear();
	 if(!xmlParser->setContent(*buf, false, &errorMsg, &errorLine, &errorColumn))
	 {
	 	timer->stop();
	 	QMessageBox::information(this, tr("XML parse error"),
                             tr("The following error occurred: %1.\nAt line %2 column %3.")
                             .arg(errorMsg).arg(errorLine).arg(errorColumn));	
        return;
 	 }
 	 QDomNodeList nodesTopologies = xmlParser->documentElement().namedItem("topologies").childNodes();
 	 QDomNodeList nodesLinks = xmlParser->documentElement().namedItem("links").childNodes();
 	 QDomNodeList neighborsLinks = xmlParser->documentElement().namedItem("neighbors").childNodes();
 	 
 	 //Read topologies from XML file and load it in a QMap for later	 
 	 for(int i=0; i<nodesTopologies.size(); i++)
 	 {
 	 	QDomNamedNodeMap args = nodesTopologies.at(i).attributes();
 	 	QString destIp = args.namedItem("destination_ip").nodeValue();
 	 	QString srcIp  = args.namedItem("last_hop_ip").nodeValue();
 	 	topologies->insertMulti(srcIp,destIp);
	 }
	 
	 //Read local ip from XML for later 	 
	 QDomNamedNodeMap argsLinks = nodesLinks.at(0).attributes();
	 localIp = argsLinks.namedItem("local_ip").nodeValue();
	 
	 //Read informations about each nodes from XML file and load it in a QMap for later	 
	 nodesInfos->clear(); 
	 nodesInfos->insert(localIp,"Local node");
 	 for(int i=0; i<nodesLinks.size(); i++)
 	 {
 	 	argsLinks = nodesLinks.at(i).attributes();
 	 	QDomNamedNodeMap argsNeighbor;
 	 	QString key = argsLinks.namedItem("remote_ip").nodeValue();
 	 	for(int j=0; j<nodesLinks.size(); j++)
 	 	{
 	 		argsNeighbor = neighborsLinks.at(j).attributes();
 	 		if(argsNeighbor.namedItem("ip").nodeValue() == key) break;
	 	}
 	 	QString infos = "";
 	 	infos.append(tr("Hyster.\t%1\n").arg(argsLinks.namedItem("hysteresis").nodeValue()));
		infos.append(tr("Link qu.\t%1\n").arg(argsLinks.namedItem("link_quality").nodeValue()));
		infos.append(tr("Lost\t%1\n").arg(argsLinks.namedItem("lost").nodeValue()));
		infos.append(tr("Total\t%1\n").arg(argsLinks.namedItem("total").nodeValue()));
		infos.append(tr("NLQ\t%1\n").arg(argsLinks.namedItem("nlq").nodeValue()));
		infos.append(tr("ETX\t%1\n").arg(argsLinks.namedItem("etx").nodeValue()));
		infos.append(tr("SYM\t%1\n").arg(argsNeighbor.namedItem("sym").nodeValue()));
 	 	infos.append(tr("MPR\t%1\n").arg(argsNeighbor.namedItem("mpr").nodeValue()));
 	 	infos.append(tr("MPRS\t%1\n").arg(argsNeighbor.namedItem("mprs").nodeValue()));
 	 	infos.append(tr("Willing.\t%1").arg(argsNeighbor.namedItem("willingness").nodeValue()));
 	 	nodesInfos->insertMulti(key,infos);
	 }
	 delete xmlParser;
	 xmlParser = NULL;
	 
	 this->repaint();
	 
}


/*
  Draw the graph from information 
*/
void OlsrdInfo::paintEvent(QPaintEvent*)
{  
	float angle;
	int radius;
    nodeSize = 12;	 
	
	//Config the painter
	QPainter p(this);
	QFontMetrics fm(p.font());
	QRect dim = p.window();
	p.translate(dim.width()/2, dim.height()/2);
	p.setRenderHint(QPainter::Antialiasing);
	radius = 0.8*qMin(dim.width(),dim.height())/2;
	
	//Foreach node compute the position
	posNodes->clear();
	angle = (M_PI * 2) / nodesInfos->size();
	int i = 0;
	QMapIterator<QString, QString> n(*nodesInfos);
	while(n.hasNext())
	{
		n.next();
		posNodes->insert(n.key(),QPointF(cos(angle*i+anim)*radius,sin(angle*i+anim)*radius));
		i++;
	}
		
	
	//Draw nodes and links between them 
	QMapIterator<QString, QPointF> node(*posNodes);
	while(node.hasNext())
	{
		node.next();
		
		//Draw nodes
		if(localIp == node.key())
			p.setBrush(Qt::red);
		else
			p.setBrush(Qt::blue);				
		p.drawEllipse(node.value().x()-nodeSize/2,node.value().y()-nodeSize/2,nodeSize,nodeSize);
		p.drawText(node.value().x()*1.5-fm.width(node.key())/2,
		           node.value().y()*1.15+fm.height()/4,
		           node.key());
		
		//Draw links
		QList<QString> links = topologies->values(node.key());
		for(int j=0; j<links.size(); j++)
		{
			QPointF destPos = posNodes->value(links.at(j));
			if((destPos.x() != 0.0) || (destPos.y() != 0.0))
				drawArrow(&p,node.value(), destPos, 8, nodeSize/2);
		}
	}
	
	//We dispaly informations about one selected node
	//It's better to use toolTip but it doesn't work wit Qtopia...
	p.drawText(-dim.width()/2+5, -dim.height()/2+5, dim.width(), dim.height(), 
				Qt::AlignLeft | Qt::TextWordWrap /*| Qt::TextExpandTabs*/, selNode+"\n" + nodesInfos->value(selNode));

}

/*
  An error occure with the socket
  We display this error
*/
void OlsrdInfo::socket_displayError(QAbstractSocket::SocketError socketError)
{
    if (socketError == QTcpSocket::RemoteHostClosedError)
        return;

	timer->stop();
    QMessageBox::information(this, tr("Network error"),
                             tr("The following error occurred: %1.")
                             .arg(tcpClient.errorString()));

    tcpClient.close();
}

/*
  When th mouse move we check if the cursor is over a node
  if true we dispaly some information about this node
*/
void OlsrdInfo::mouseMoveEvent(QMouseEvent * event)
{
	selNode = "";
	int x = event->x()-width()/2;
	int y = event->y()-height()/2;
	setWindowTitle(tr("QOlsd - Qt Gui for OLSD - ReDS"));
	QMapIterator<QString, QPointF> node(*posNodes);
	while(node.hasNext() && selNode.isEmpty())
	{
		node.next();
		int dx = x-node.value().x();
		int dy = y-node.value().y();
		if((dx*dx+dy*dy) <= (nodeSize*nodeSize))
			selNode = node.key();		
	}
}
